package dev.pfaff.glass.glfw;

import dev.pfaff.glass.glfw.callbacks.ErrorCallback;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.nio.charset.StandardCharsets;

import static dev.pfaff.glass.glfw.GLFWwindow.*;
import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodType.methodType;

public final class GLFW implements AutoCloseable {
	static final Linker LINKER = Linker.nativeLinker();

	private static final MethodHandle MH_errorCallbackImpl;
	public static final FunctionDescriptor FD_GLFWerrorfun = FunctionDescriptor.ofVoid(JAVA_INT,
			ADDRESS.withTargetLayout(JAVA_BYTE));

	private static Arena errorCallbackArena;

	static {
		try {
			MH_errorCallbackImpl = MethodHandles.lookup().findStatic(GLFW.class,
					"errorCallbackImpl",
					methodType(void.class, ErrorCallback.class, int.class, MemorySegment.class));
		} catch (NoSuchMethodException | IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	private static void errorCallbackImpl(ErrorCallback callback, int code, MemorySegment descriptionPtr) {
		var description = descriptionPtr.reinterpret(Long.MAX_VALUE).getString(0, StandardCharsets.UTF_8);
		callback.errorOccurred(code, description);
	}

	public static final AddressLayout ML_GLFWwindow = ADDRESS.withName("GLFWwindow*");
	public static final AddressLayout ML_GLFWmonitor = ADDRESS.withName("GLFWmonitor*");
	public static final AddressLayout ML_GLFWerrorfun = ADDRESS.withName("GLFWerrorfun");

	public static final MethodHandle MH_glfwInit = LINKER.downcallHandle(FunctionDescriptor.of(JAVA_BOOLEAN));
	public static final MethodHandle MH_glfwTerminate = LINKER.downcallHandle(FunctionDescriptor.ofVoid());
	public static final MethodHandle MH_glfwPollEvents = LINKER.downcallHandle(FunctionDescriptor.ofVoid());
	public static final MethodHandle MH_glfwSetErrorCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWerrorfun,
			ML_GLFWerrorfun));

	/**
	 * The thread that GLFW was initialized on.
	 */
	private final Thread mainThread = Thread.currentThread();

	private final MemorySegment MS_glfwInit;
	private final MemorySegment MS_glfwTerminate;
	private final MemorySegment MS_glfwPollEvents;
	private final MemorySegment MS_glfwSetErrorCallback;

	private final MemorySegment MS_glfwMakeContextCurrent;
	private final MemorySegment MS_glfwGetCurrentContext;
	private final MemorySegment MS_glfwSwapInterval;
	private final MemorySegment MS_glfwExtensionSupported;
	private final MemorySegment MS_glfwCreateWindow;
	private final MemorySegment MS_glfwWindowHint;
	private final MemorySegment MS_glfwDestroyWindow;
	private final MemorySegment MS_glfwWindowShouldClose;
	private final MemorySegment MS_glfwGetFramebufferSize;
	private final MemorySegment MS_glfwGetWindowSize;
	private final MemorySegment MS_glfwGetWindowContentScale;
	private final MemorySegment MS_glfwGetWindowAttrib;
	private final MemorySegment MS_glfwSwapBuffers;
	private final MemorySegment MS_glfwShowWindow;
	private final MemorySegment MS_glfwHideWindow;
	private final MemorySegment MS_glfwSetWindowSize;
	private final MemorySegment MS_glfwGetClipboardString;
	private final MemorySegment MS_glfwSetClipboardString;

	final MemorySegment MS_glfwSetCharCallback;
	final MemorySegment MS_glfwSetCursorPosCallback;
	final MemorySegment MS_glfwSetKeyCallback;
	final MemorySegment MS_glfwSetMouseButtonCallback;
	final MemorySegment MS_glfwSetDropCallback;
	final MemorySegment MS_glfwSetScrollCallback;
	final MemorySegment MS_glfwSetWindowPosCallback;
	final MemorySegment MS_glfwSetWindowSizeCallback;
	final MemorySegment MS_glfwSetWindowCloseCallback;
	final MemorySegment MS_glfwSetWindowRefreshCallback;
	final MemorySegment MS_glfwSetWindowFocusCallback;
	final MemorySegment MS_glfwSetWindowIconifyCallback;
	final MemorySegment MS_glfwSetWindowMaximizeCallback;
	final MemorySegment MS_glfwSetFramebufferSizeCallback;
	final MemorySegment MS_glfwSetWindowContentScaleCallback;

	// bound for performance
	private final MethodHandle MH_glfwPollEvents_bound;

	final MethodHandle MH_glfwMakeContextCurrent_bound;
	final MethodHandle MH_glfwGetCurrentContext_bound;
	final MethodHandle MH_glfwSwapInterval_bound;
	final MethodHandle MH_glfwExtensionSupported_bound;
	final MethodHandle MH_glfwCreateWindow_bound;
	final MethodHandle MH_glfwWindowHint_bound;
	final MethodHandle MH_glfwDestroyWindow_bound;
	final MethodHandle MH_glfwWindowShouldClose_bound;
	final MethodHandle MH_glfwGetFramebufferSize_bound;
	final MethodHandle MH_glfwGetWindowSize_bound;
	final MethodHandle MH_glfwGetWindowContentScale_bound;
	final MethodHandle MH_glfwGetWindowAttrib_bound;
	final MethodHandle MH_glfwSwapBuffers_bound;
	final MethodHandle MH_glfwShowWindow_bound;
	final MethodHandle MH_glfwHideWindow_bound;
	final MethodHandle MH_glfwSetWindowSize_bound;
	final MethodHandle MH_glfwGetClipboardString_bound;
	final MethodHandle MH_glfwSetClipboardString_bound;

	public GLFW(SymbolLookup symbolLookup) {
		this.MS_glfwInit = symbolLookup.find("glfwInit").get();
		this.MS_glfwTerminate = symbolLookup.find("glfwTerminate").get();
		this.MS_glfwPollEvents = symbolLookup.find("glfwPollEvents").get();
		this.MS_glfwSetErrorCallback = symbolLookup.find("glfwSetErrorCallback").get();

		this.MS_glfwMakeContextCurrent = symbolLookup.find("glfwMakeContextCurrent").get();
		this.MS_glfwGetCurrentContext = symbolLookup.find("glfwGetCurrentContext").get();
		this.MS_glfwSwapInterval = symbolLookup.find("glfwSwapInterval").get();
		this.MS_glfwExtensionSupported = symbolLookup.find("glfwExtensionSupported").get();
		this.MS_glfwCreateWindow = symbolLookup.find("glfwCreateWindow").get();
		this.MS_glfwWindowHint = symbolLookup.find("glfwWindowHint").get();
		this.MS_glfwDestroyWindow = symbolLookup.find("glfwDestroyWindow").get();
		this.MS_glfwWindowShouldClose = symbolLookup.find("glfwWindowShouldClose").get();
		this.MS_glfwGetFramebufferSize = symbolLookup.find("glfwGetFramebufferSize").get();
		this.MS_glfwGetWindowSize = symbolLookup.find("glfwGetWindowSize").get();
		this.MS_glfwGetWindowContentScale = symbolLookup.find("glfwGetWindowContentScale").get();
		this.MS_glfwGetWindowAttrib = symbolLookup.find("glfwGetWindowAttrib").get();
		this.MS_glfwSwapBuffers = symbolLookup.find("glfwSwapBuffers").get();
		this.MS_glfwShowWindow = symbolLookup.find("glfwShowWindow").get();
		this.MS_glfwHideWindow = symbolLookup.find("glfwHideWindow").get();
		this.MS_glfwSetWindowSize = symbolLookup.find("glfwSetWindowSize").get();
		this.MS_glfwGetClipboardString = symbolLookup.find("glfwGetClipboardString").get();
		this.MS_glfwSetClipboardString = symbolLookup.find("glfwSetClipboardString").get();

		this.MS_glfwSetCharCallback = symbolLookup.find("glfwSetCharCallback").get();
		this.MS_glfwSetCursorPosCallback = symbolLookup.find("glfwSetCursorPosCallback").get();
		this.MS_glfwSetKeyCallback = symbolLookup.find("glfwSetKeyCallback").get();
		this.MS_glfwSetMouseButtonCallback = symbolLookup.find("glfwSetMouseButtonCallback").get();
		this.MS_glfwSetDropCallback = symbolLookup.find("glfwSetDropCallback").get();
		this.MS_glfwSetScrollCallback = symbolLookup.find("glfwSetScrollCallback").get();
		this.MS_glfwSetWindowPosCallback = symbolLookup.find("glfwSetWindowPosCallback").get();
		this.MS_glfwSetWindowSizeCallback = symbolLookup.find("glfwSetWindowSizeCallback").get();
		this.MS_glfwSetWindowCloseCallback = symbolLookup.find("glfwSetWindowCloseCallback").get();
		this.MS_glfwSetWindowRefreshCallback = symbolLookup.find("glfwSetWindowRefreshCallback").get();
		this.MS_glfwSetWindowFocusCallback = symbolLookup.find("glfwSetWindowFocusCallback").get();
		this.MS_glfwSetWindowIconifyCallback = symbolLookup.find("glfwSetWindowIconifyCallback").get();
		this.MS_glfwSetWindowMaximizeCallback = symbolLookup.find("glfwSetWindowMaximizeCallback").get();
		this.MS_glfwSetFramebufferSizeCallback = symbolLookup.find("glfwSetFramebufferSizeCallback").get();
		this.MS_glfwSetWindowContentScaleCallback = symbolLookup.find("glfwSetWindowContentScaleCallback").get();

		this.MH_glfwPollEvents_bound = MH_glfwPollEvents.bindTo(MS_glfwPollEvents);

		this.MH_glfwMakeContextCurrent_bound = MH_glfwMakeContextCurrent.bindTo(MS_glfwMakeContextCurrent);
		this.MH_glfwGetCurrentContext_bound = MH_glfwGetCurrentContext.bindTo(MS_glfwGetCurrentContext);
		this.MH_glfwSwapInterval_bound = MH_glfwSwapInterval.bindTo(MS_glfwSwapInterval);
		this.MH_glfwExtensionSupported_bound = MH_glfwExtensionSupported.bindTo(MS_glfwExtensionSupported);
		this.MH_glfwCreateWindow_bound = MH_glfwCreateWindow.bindTo(MS_glfwCreateWindow);
		this.MH_glfwWindowHint_bound = MH_glfwWindowHint.bindTo(MS_glfwWindowHint);
		this.MH_glfwDestroyWindow_bound = MH_glfwDestroyWindow.bindTo(MS_glfwDestroyWindow);
		this.MH_glfwWindowShouldClose_bound = MH_glfwWindowShouldClose.bindTo(MS_glfwWindowShouldClose);
		this.MH_glfwGetFramebufferSize_bound = MH_glfwGetFramebufferSize.bindTo(MS_glfwGetFramebufferSize);
		this.MH_glfwGetWindowSize_bound = MH_glfwGetWindowSize.bindTo(MS_glfwGetWindowSize);
		this.MH_glfwGetWindowContentScale_bound = MH_glfwGetWindowContentScale.bindTo(MS_glfwGetWindowContentScale);
		this.MH_glfwGetWindowAttrib_bound = MH_glfwGetWindowAttrib.bindTo(MS_glfwGetWindowAttrib);
		this.MH_glfwSwapBuffers_bound = MH_glfwSwapBuffers.bindTo(MS_glfwSwapBuffers);
		this.MH_glfwShowWindow_bound = MH_glfwShowWindow.bindTo(MS_glfwShowWindow);
		this.MH_glfwHideWindow_bound = MH_glfwHideWindow.bindTo(MS_glfwHideWindow);
		this.MH_glfwSetWindowSize_bound = MH_glfwSetWindowSize.bindTo(MS_glfwSetWindowSize);
		this.MH_glfwGetClipboardString_bound = MH_glfwGetClipboardString.bindTo(MS_glfwGetClipboardString);
		this.MH_glfwSetClipboardString_bound = MH_glfwSetClipboardString.bindTo(MS_glfwSetClipboardString);

		// initialized GLFW
		boolean result;
		try {
			result = (boolean) MH_glfwInit.invokeExact(MS_glfwInit);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (!result) {
			throw new RuntimeException("Failed to initialize GLFW");
		}
	}

	// TODO: error callback

	/**
	 * Deinitializes GLFW.
	 */
	@Override
	public void close() {
		// technically this and init can be called multiple times, but supporting that while maintaining thread safety
		// checks would significantly complicate the design.
		try {
			MH_glfwTerminate.invokeExact(MS_glfwTerminate);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void requireOnMainThread() {
		if (Thread.currentThread() != mainThread) {
			throw new IllegalStateException("Must be on main thread!");
		}
	}

	public void pollEvents() {
		try {
			MH_glfwPollEvents_bound.invokeExact();
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	static <T> Arena setCallback(@NonNull MethodHandle MH_setCallback,
								 @NonNull MemorySegment MS_setCallback,
								 @NonNull T callback,
								 @NonNull MethodHandle invokerStub,
								 @NonNull FunctionDescriptor invokerFunctionDescriptor,
								 @Nullable Arena oldArena) {
		Arena newArena = Arena.ofShared();
		try {
			var stub = LINKER.upcallStub(invokerStub.bindTo(callback), invokerFunctionDescriptor, newArena);
			var _ = (MemorySegment) MH_setCallback.invokeExact(MS_setCallback, stub);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (oldArena != null) oldArena.close();
		return newArena;
	}

	public void setErrorCallback(ErrorCallback callback) {
		errorCallbackArena = setCallback(MH_glfwSetErrorCallback,
				MS_glfwSetErrorCallback,
				callback,
				MH_errorCallbackImpl,
				FD_GLFWerrorfun,
				errorCallbackArena);
	}
}
